/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TEncAdaptiveLoopFilter.cpp
 \brief    estimation part of adaptive loop filter class
 */
#include "TEncLoopFilter.h"
#include <math.h>

//! \ingroup TLibEncoder
//! \{
#if ADAPT_DF_OFFSET
// ====================================================================================================================
// Constants
// ====================================================================================================================

#define   EDGE_VER    0
#define   EDGE_HOR    1
#define   QpUV(iQpY)  ( g_aucChromaScale[ max( min( (iQpY), MAX_QP ), MIN_QP ) ] )


#if (PARALLEL_DEBLK_DECISION && !PARALLEL_MERGED_DEBLK)
#define   DECIDE_FILTER                0
#define   EXECUTE_FILTER               1
#define   DECIDE_AND_EXECUTE_FILTER    2
#endif

#define F143 1
#if F143
#define DEFAULT_INTRA_TC_OFFSET 2 ///< Default intra TC offset
#else
#define DEFAULT_INTRA_TC_OFFSET 4 ///< Default intra TC offset
#endif


// ====================================================================================================================
// Tables
// ====================================================================================================================

const UChar tctable_8x8[56] =
{
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,1,1,1,1,1,1,1,1,1,2,2,2,2,3,3,3,3,4,4,4,5,5,6,6,7,8,9,9,10,10,11,11,12,12,13,13,14,14
};

const UChar betatable_8x8[52] =
{
  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,6,7,8,9,10,11,12,13,14,15,16,17,18,20,22,24,26,28,30,32,34,36,38,40,42,44,46,48,50,52,54,56,58,60,62,64
};

// ====================================================================================================================
// Constructor / destructor
// ====================================================================================================================
TEncLoopFilter::TEncLoopFilter()
//: m_uiNumPartitions( 0 )
{
}

TEncLoopFilter::~TEncLoopFilter()
{
}

Void TEncLoopFilter::startFindLoopFilterParamsEnc( TEncEntropy* pcEntropyCoder )
{
  m_pcEntropyCoder = pcEntropyCoder;  
}

Void TEncLoopFilter::endFindLoopFilterParamsEnc()
{
  m_pcEntropyCoder = NULL;
}



Void TEncLoopFilter::findLoopFilterParamsFast( TComPic* pcPic, Int BetaRange, Int TcRange, Int SideBetaRange )
{
  m_iBetaRange = Clip3(0,MAX_OFFSET,abs(BetaRange));
  m_iTcRange =   Clip3(0,MAX_OFFSET,abs(TcRange));
  m_iSideBetaRange = Clip3(0,MAX_OFFSET,abs(SideBetaRange));

  m_iBestBeta = 0;           
  m_iBestSideBeta = 0;
  m_iBestTc = 0;

  m_iBestBetaOffset = 0;           
  m_iBestSideBetaOffset = 0;
  m_iBestTcOffset = 0;


  m_pcPic = pcPic;


  m_uiOptPass = 1;

  loopFilterEstimatePic( pcPic );

  UInt64 SSDCur = 0;


  Double dLambda = pcPic->getSlice(0)->getLambdaLuma();
  UInt64 uiBits = 1;
  Double bestCost = MAX_DOUBLE;
  Double curCost;


  for(Int iCurBetaOffset = -m_iBetaRange;  iCurBetaOffset< m_iBetaRange + 1 ; iCurBetaOffset++)
  {
    for(Int iCurSideBetaOffset = -m_iSideBetaRange;  iCurSideBetaOffset< m_iSideBetaRange +1 ; iCurSideBetaOffset++)
    {
      m_pcEntropyCoder->resetEntropy();
      m_pcEntropyCoder->resetBits();         
      pcPic->getSlice(0)->setLoopFilterBetaOffset(findBetaOffset(iCurBetaOffset,pcPic->getSlice(0)->getSliceQp(),1));  
      pcPic->getSlice(0)->setLoopFilterTcOffset(0);
      pcPic->getSlice(0)->setLoopFilterSideBetaOffset(findSideBetaOffset(iCurSideBetaOffset,pcPic->getSlice(0)->getSliceQp(),4));
      m_pcEntropyCoder->encodeDFParamsSlice( pcPic->getSlice(0));
      uiBits = m_pcEntropyCoder->getNumberOfWrittenBits();
      SSDCur = m_iCounter[MAX_OFFSET+iCurBetaOffset][MAX_OFFSET+iCurSideBetaOffset];
      curCost = (((Double)SSDCur) + ((Double)uiBits * dLambda));
      if (curCost < bestCost ||  (curCost == bestCost && ( (abs(iCurBetaOffset)<=abs(m_iBestBeta) )&&(abs(iCurSideBetaOffset)<=abs(m_iBestSideBeta) ) ) ))
      {
        bestCost = curCost;
        m_iBestBeta = iCurBetaOffset;
        m_iBestBetaOffset = findBetaOffset(iCurBetaOffset,pcPic->getSlice(0)->getSliceQp(),1);           
        m_iBestBeta = iCurSideBetaOffset;
        m_iBestSideBetaOffset = findSideBetaOffset(iCurSideBetaOffset,pcPic->getSlice(0)->getSliceQp(),4);

      }
    }
  }

  setCfg( pcPic->getSlice(0)->getLoopFilterDisable(), m_iBestBetaOffset, 0, m_iBestSideBetaOffset);
  pcPic->getSlice(0)->setLoopFilterBetaOffset(m_iBestBetaOffset);
  pcPic->getSlice(0)->setLoopFilterTcOffset(0);
  pcPic->getSlice(0)->setLoopFilterSideBetaOffset(m_iBestSideBetaOffset);

  m_uiOptPass = 2;

  loopFilterEstimatePic( pcPic );



  uiBits = 1;
  bestCost = MAX_DOUBLE;


  for(Int iCurTcOffset = -m_iTcRange;  iCurTcOffset< m_iTcRange + 1 ; iCurTcOffset++)
  {
    m_pcEntropyCoder->resetEntropy();
    m_pcEntropyCoder->resetBits();

    pcPic->getSlice(0)->setLoopFilterTcOffset(findTcOffset(iCurTcOffset,pcPic->getSlice(0)->getSliceQp(),2));

    m_pcEntropyCoder->encodeDFParamsSlice( pcPic->getSlice(0));
    uiBits = m_pcEntropyCoder->getNumberOfWrittenBits();
    SSDCur = m_iCounter[MAX_OFFSET+iCurTcOffset][0];
    curCost = (((Double)SSDCur) + ((Double)uiBits * dLambda));
    if (curCost < bestCost ||  (curCost == bestCost && ( (abs(iCurTcOffset)<=abs(m_iBestTc) ) ) ))
    {
      bestCost = curCost;

      m_iBestTc = iCurTcOffset;    
      m_iBestTcOffset = findTcOffset(iCurTcOffset,pcPic->getSlice(0)->getSliceQp(),2);

    }
  }

  pcPic->getSlice(0)->setLoopFilterTcOffset(m_iBestTcOffset);
  setCfg( pcPic->getSlice(0)->getLoopFilterDisable(), m_iBestBetaOffset, m_iBestTcOffset, m_iBestSideBetaOffset);

}

Void TEncLoopFilter::loopFilterEstimatePic( TComPic* pcPic )
{

  ::memset( m_iCounter, 0, sizeof( UInt64) * (MAX_OFFSET*2+1) * (MAX_OFFSET*2+1) );

#if PARALLEL_MERGED_DEBLK

  // Horizontal filtering
  for ( UInt uiCUAddr = 0; uiCUAddr < pcPic->getNumCUsInFrame(); uiCUAddr++ )
  {
    TComDataCU* pcCU = pcPic->getCU( uiCUAddr );

#if DEBLK_CLEANUP_CHROMA_BS
    ::memset( m_aapucBS       [EDGE_VER], 0, sizeof( UChar ) * m_uiNumPartitions );
#endif
    for( Int iPlane = 0; iPlane < 3; iPlane++ )
    {
#if !DEBLK_CLEANUP_CHROMA_BS
      ::memset( m_aapucBS       [EDGE_VER][iPlane], 0, sizeof( UChar ) * m_uiNumPartitions );
#endif
      ::memset( m_aapbEdgeFilter[EDGE_VER][iPlane], 0, sizeof( bool  ) * m_uiNumPartitions );
    }

    // CU-based deblocking
    xEstimateCU( pcCU, 0, 0, EDGE_VER );
  }

  // Vertical filtering
  for ( UInt uiCUAddr = 0; uiCUAddr < pcPic->getNumCUsInFrame(); uiCUAddr++ )
  {
    TComDataCU* pcCU = pcPic->getCU( uiCUAddr );

#if DEBLK_CLEANUP_CHROMA_BS
      ::memset( m_aapucBS       [EDGE_HOR], 0, sizeof( UChar ) * m_uiNumPartitions );
#endif
    for( Int iPlane = 0; iPlane < 3; iPlane++ )
    {
#if !DEBLK_CLEANUP_CHROMA_BS
      ::memset( m_aapucBS       [EDGE_HOR][iPlane], 0, sizeof( UChar ) * m_uiNumPartitions );
#endif
      ::memset( m_aapbEdgeFilter[EDGE_HOR][iPlane], 0, sizeof( bool  ) * m_uiNumPartitions );
    }
    // CU-based deblocking
    xEstimateCU( pcCU, 0, 0, EDGE_HOR );
  }
#else
  // for every CU
  for ( UInt uiCUAddr = 0; uiCUAddr < pcPic->getNumCUsInFrame(); uiCUAddr++ )
  {
    TComDataCU* pcCU = pcPic->getCU( uiCUAddr );
    
    for( Int iDir = EDGE_VER; iDir <= EDGE_HOR; iDir++ )
    {
      for( Int iPlane = 0; iPlane < 3; iPlane++ )
      {
        ::memset( m_aapucBS       [iDir][iPlane], 0, sizeof( UChar ) * m_uiNumPartitions );
        ::memset( m_aapbEdgeFilter[iDir][iPlane], 0, sizeof( bool  ) * m_uiNumPartitions );
      }
    }
    // CU-based deblocking
    xEstimateCU( pcCU, 0, 0 );
  }
#endif
  

}


// ====================================================================================================================
// Protected member functions
// ====================================================================================================================

#if PARALLEL_MERGED_DEBLK
/**
 - Deblocking filter process in CU-based (the same function as conventional's)
 .
 \param Edge          the direction of the edge in block boundary (horizonta/vertical), which is added newly
*/
Void TEncLoopFilter::xEstimateCU( TComDataCU* pcCU, UInt uiAbsZorderIdx, UInt uiDepth, Int Edge )
#else
Void TEncLoopFilter::xEstimateCU( TComDataCU* pcCU, UInt uiAbsZorderIdx, UInt uiDepth )
#endif
{
  TComPic* pcPic     = pcCU->getPic();
  UInt uiCurNumParts = pcPic->getNumPartInCU() >> (uiDepth<<1);
  UInt uiQNumParts   = uiCurNumParts>>2;
  
  if( pcCU->getDepth(uiAbsZorderIdx) > uiDepth )
  {
    for ( UInt uiPartIdx = 0; uiPartIdx < 4; uiPartIdx++, uiAbsZorderIdx+=uiQNumParts )
    {
      UInt uiLPelX   = pcCU->getCUPelX() + g_auiRasterToPelX[ g_auiZscanToRaster[uiAbsZorderIdx] ];
      UInt uiTPelY   = pcCU->getCUPelY() + g_auiRasterToPelY[ g_auiZscanToRaster[uiAbsZorderIdx] ];
      if( ( uiLPelX < pcCU->getSlice()->getSPS()->getWidth() ) && ( uiTPelY < pcCU->getSlice()->getSPS()->getHeight() ) )
#if PARALLEL_MERGED_DEBLK
        xEstimateCU( pcCU, uiAbsZorderIdx, uiDepth+1, Edge );
#else
        xEstimateCU( pcCU, uiAbsZorderIdx, uiDepth+1 );
#endif
    }
    return;
  }
  
  xSetLoopfilterParam( pcCU, uiAbsZorderIdx );
  
#if NSQT
  xSetEdgefilterTU   ( pcCU, g_auiZscanToRaster[ uiAbsZorderIdx ], uiAbsZorderIdx, uiDepth );
#else
  xSetEdgefilterTU   ( pcCU, uiAbsZorderIdx, uiDepth );
#endif
  xSetEdgefilterPU   ( pcCU, uiAbsZorderIdx );
  
#if PARALLEL_MERGED_DEBLK
  Int iDir = Edge;
#else
  for ( Int iDir = EDGE_VER; iDir <= EDGE_HOR; iDir++ )
#endif
  {
    for( UInt uiPartIdx = uiAbsZorderIdx; uiPartIdx < uiAbsZorderIdx + uiCurNumParts; uiPartIdx++ )
    {
      if ( m_aapbEdgeFilter[iDir][0][uiPartIdx] )
      {
        xGetBoundaryStrengthSingle ( pcCU, uiAbsZorderIdx, iDir, uiPartIdx );
      }
    }
  }
  
  UInt uiPelsInPart = g_uiMaxCUWidth >> g_uiMaxCUDepth;
  UInt PartIdxIncr = DEBLOCK_SMALLEST_BLOCK / uiPelsInPart ? DEBLOCK_SMALLEST_BLOCK / uiPelsInPart : 1 ;
  
  UInt uiSizeInPU = pcPic->getNumPartInWidth()>>(uiDepth);
  

#if !PARALLEL_MERGED_DEBLK
  for ( Int iDir = EDGE_VER; iDir <= EDGE_HOR; iDir++ )
#endif
  {
    for ( UInt iEdge = 0; iEdge < uiSizeInPU ; iEdge+=PartIdxIncr)
    {
      xEdgeEstimateLuma     ( pcCU, uiAbsZorderIdx, uiDepth, iDir, iEdge );
    }
  }
}



Void TEncLoopFilter::xEdgeEstimateLuma( TComDataCU* pcCU, UInt uiAbsZorderIdx, UInt uiDepth, Int iDir, Int iEdge  )
{
  TComPicYuv* pcPicYuvRec = pcCU->getPic()->getPicYuvRec();
  Pel* piSrc    = pcPicYuvRec->getLumaAddr( pcCU->getAddr(), uiAbsZorderIdx );
  Pel* piTmpSrc = piSrc;
  Pel* piSrcJudge    = pcCU->getPic()->getPicYuvOrg()->getLumaAddr( pcCU->getAddr(), uiAbsZorderIdx );
  Pel* piTmpSrcJudge = piSrcJudge;

  Int  iStride = pcPicYuvRec->getStride();
  Int  iQP = pcCU->getQP( uiAbsZorderIdx );

  UInt uiNumParts = pcCU->getPic()->getNumPartInWidth()>>uiDepth;
  
  UInt  uiPelsInPart = g_uiMaxCUWidth >> g_uiMaxCUDepth;
  UInt  PartIdxIncr = DEBLOCK_SMALLEST_BLOCK / uiPelsInPart ? DEBLOCK_SMALLEST_BLOCK / uiPelsInPart : 1;
  UInt  uiBlocksInPart = uiPelsInPart / DEBLOCK_SMALLEST_BLOCK ? uiPelsInPart / DEBLOCK_SMALLEST_BLOCK : 1;
  UInt  uiBsAbsIdx, uiBs;
  Int   iOffset, iSrcStep;

  if (iDir == EDGE_VER)
  {
    iOffset = 1;
    iSrcStep = iStride;
    piTmpSrc += iEdge*uiPelsInPart;
    piTmpSrcJudge += iEdge*uiPelsInPart;
  }
  else  // (iDir == EDGE_HOR)
  {
    iOffset = iStride;
    iSrcStep = 1;
    piTmpSrc += iEdge*uiPelsInPart*iStride;
    piTmpSrcJudge += iEdge*uiPelsInPart*iStride;
  }
  
  for ( UInt iIdx = 0; iIdx < uiNumParts; iIdx+=PartIdxIncr )
  {
    
    uiBs = 0;
    for (UInt iIdxInside = 0; iIdxInside<PartIdxIncr; iIdxInside++)
    {
      uiBsAbsIdx = xCalcBsIdx( pcCU, uiAbsZorderIdx, iDir, iEdge, iIdx+iIdxInside);
      if (uiBs < m_aapucBS[iDir][uiBsAbsIdx])
      {
        uiBs = m_aapucBS[iDir][uiBsAbsIdx];
      }
    }
    
    Int iBitdepthScale = (1<<(g_uiBitIncrement+g_uiBitDepth-8));
    
    Int uiTcOffset = ( uiBs > 1 ) ? DEFAULT_INTRA_TC_OFFSET : 0;

    for (UInt iBlkIdx = 0; iBlkIdx < uiBlocksInPart*2; iBlkIdx += 2)
    {
      if ( uiBs )
      {
        if(m_uiOptPass == 1)
        {
          estimateBlock(piTmpSrc, piTmpSrcJudge, iSrcStep, iIdx, uiPelsInPart, iBlkIdx, iOffset, uiTcOffset, iBitdepthScale, iQP  );
          estimateBlock(piTmpSrc, piTmpSrcJudge, iSrcStep, iIdx, uiPelsInPart, iBlkIdx+1, iOffset, uiTcOffset, iBitdepthScale, iQP  );
        }

        if(m_uiOptPass == 2)
        {
          estimateTc(piTmpSrc,  piTmpSrcJudge, iSrcStep, iIdx, uiPelsInPart, iBlkIdx, iOffset,  uiTcOffset, iBitdepthScale, iQP  );
          estimateTc(piTmpSrc,  piTmpSrcJudge, iSrcStep, iIdx, uiPelsInPart, iBlkIdx+1, iOffset,  uiTcOffset, iBitdepthScale, iQP  );
        }      
      }
    }
  }
}


Void TEncLoopFilter::estimateBlock(Pel* piSrc, Pel* piOrig, Int iSrcStep, Int iIdx, UInt uiPelsInPart, Int iBlkIdx, Int iOffset, UInt uiTcOffset, Int iBitdepthScale, Int iQP )
{

  Int dp0 = xCalcDP( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+0), iOffset);
  Int dq0 = xCalcDQ( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+0), iOffset);

  Int dp3 = xCalcDP( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+3), iOffset);
  Int dq3 = xCalcDQ( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+3), iOffset);

  Int d0 = dp0 + dq0;
  Int d3 = dp3 + dq3;
  Int dp = dp0 + dp3;
  Int dq = dq0 + dq3;
  Int d = dp + dq;

  Int iDStrong0 = xCalcDStrong( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+0), iOffset); 
  Int iDStrong3 = xCalcDStrong( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+3), iOffset);

  Int iStrongMiddle0 = xCalcStrongMiddle( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+0), iOffset); 
  Int iStrongMiddle3 = xCalcStrongMiddle( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+3), iOffset);

  Int iNaturalEdge0 = xCalcThrCut( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+0), iOffset);
  Int iNaturalEdge3 = xCalcThrCut( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+3), iOffset);

  Int iIndexTC = Clip3(0, MAX_QP+2, (iQP + (Int)uiTcOffset) );
  Int iIndexB = Clip3(0, MAX_QP, iQP );

  Int iTc =  tctable_8x8[iIndexTC]*iBitdepthScale;

  Int iBeta = betatable_8x8[iIndexB]*iBitdepthScale;

  Int iSideThreshold = ((iBeta+(iBeta>>1))>>3);
  Int iThrCut = iTc*10;

  UInt uiSSDCenterLine = 0, uiSSD2PelPLine = 0, uiSSD2PelQLine = 0 ;   
  m_uiSSDStrong = 0, m_uiSSDWeak = 0, m_uiSSDOrig = 0, m_uiSSDCenter = 0, m_uiSSD2PelP = 0, m_uiSSD2PelQ = 0; 


  for ( UInt i = 0; i < DEBLOCK_SMALLEST_BLOCK/2; i++)
  {
    m_uiSSDOrig += xSSDPelOrig( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+i), piOrig+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+i), iOffset );
    m_uiSSDStrong += xSSDPelStrongFilt( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+i), piOrig+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+i), iOffset );
    m_uiSSDWeak += xSSDPelWeakFilt(  piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+i), piOrig+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+i), iOffset, uiSSD2PelPLine, uiSSD2PelQLine, uiSSDCenterLine, iTc);
    m_uiSSDCenter += uiSSDCenterLine; 
    m_uiSSD2PelP += uiSSD2PelPLine; 
    m_uiSSD2PelQ += uiSSD2PelQLine; 
  }

  UInt SSDCur;
  Int iBestBetaOffset, iBestTcOffset, iBestSideBetaOffset, iBestThrCutOffset;
  iBestBetaOffset = iBestTcOffset = iBestSideBetaOffset = iBestThrCutOffset = 0;


  for(Int iCurBetaOffset = -m_iBetaRange;  iCurBetaOffset< m_iBetaRange + 1 ; iCurBetaOffset++)
  {
    Int iBetaV = Clip3(0,64*iBitdepthScale, (iBeta+iCurBetaOffset*iBitdepthScale));

    for(Int iCurSideBetaOffset = -m_iSideBetaRange;  iCurSideBetaOffset< m_iSideBetaRange +1 ; iCurSideBetaOffset++)
    {
      Int iSideThresholdV = Clip3(0,14*iBitdepthScale, (iSideThreshold + iCurSideBetaOffset*iBitdepthScale));


      SSDCur = xCalcSSDFilterEdge(iBetaV, iTc, iThrCut, iSideThresholdV, d, d0, d3, dp, dq, iNaturalEdge0, iNaturalEdge3, iStrongMiddle0, iStrongMiddle3, iDStrong0, iDStrong3); 

      m_iCounter[MAX_OFFSET+iCurBetaOffset][MAX_OFFSET+iCurSideBetaOffset] += SSDCur;       
    }
  }
}



__inline UInt TEncLoopFilter::xCalcSSDFilterEdge(Int beta, Int tc, Int iThrCut, Int iSideThr, Int d, Int d0, Int d3, Int dp, Int dq, Int iNaturalEdge0, Int iNaturalEdge3, Int iStrongMiddle0, Int iStrongMiddle3, Int iDStrong0, Int iDStrong3)
{
  if( d >= beta)
  {
    return m_uiSSDOrig;
  }

  if ( (iDStrong0 < (beta>>3)) && (2*d0<(beta>>2)) && ( iStrongMiddle0 < ((tc*5+1)>>1)) && (iDStrong3 < (beta>>3)) && (2*d3<(beta>>2)) && ( iStrongMiddle3 < ((tc*5+1)>>1)) ) //strong filtering
  {
    return m_uiSSDStrong;
  }

  if ( (iNaturalEdge0 < iThrCut) && (iNaturalEdge3 < iThrCut) )
  {
    if(dp < iSideThr)
    {
      if(dq < iSideThr)
      {
        return m_uiSSDWeak;
      }
      else
      {
        return m_uiSSD2PelP;
      }
    }
    if ( dq < iSideThr )
    {
      return m_uiSSD2PelQ;
    }

    return m_uiSSDCenter;
  }

  return m_uiSSDOrig;
}




Void TEncLoopFilter::estimateTc(Pel* piSrc, Pel* piOrig, Int iSrcStep, Int iIdx, UInt uiPelsInPart, Int iBlkIdx, Int iOffset, UInt uiTcOffset, Int iBitdepthScale, Int iQP )
{
  Int dp0 = xCalcDP( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+0), iOffset);
  Int dq0 = xCalcDQ( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+0), iOffset);

  Int dp3 = xCalcDP( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+3), iOffset);
  Int dq3 = xCalcDQ( piSrc+iSrcStep*(iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2+3), iOffset);

  Int d0 = dp0 + dq0;
  Int d3 = dp3 + dq3;
  Int dp = dp0 + dp3;
  Int dq = dq0 + dq3;
  Int d = dp + dq;

  Int iIndexTC = Clip3(0, MAX_QP+2, (iQP + (Int)uiTcOffset) );
  Int iIndexB = Clip3(0, MAX_QP, iQP + (m_iBestBetaOffset ));
  Int iIndexSideB = Clip3(0, MAX_QP, iQP + (m_iBestSideBetaOffset << 2));


  Int iTc =  tctable_8x8[iIndexTC]*iBitdepthScale;

  Int iBeta = betatable_8x8[iIndexB]*iBitdepthScale;
  Int iForSideBeta = betatable_8x8[iIndexSideB]*iBitdepthScale;
  Int iSideThreshold = (iForSideBeta+(iForSideBeta>>1))>>3;
  Int iThrCut = 10 * iTc;

  xCalcSSDFilterEdgeRealTc(piSrc, iSrcStep, iOffset, iIdx*uiPelsInPart+iBlkIdx*DEBLOCK_SMALLEST_BLOCK/2, piOrig, d, d0, d3, dp, dq, iBeta, iTc, iThrCut, iSideThreshold, iBitdepthScale);

}




__inline Void TEncLoopFilter::xCalcSSDFilterEdgeRealTc(Pel* piTmpSrc, Int iSrcStep, Int iOffset, Int iShift, Pel* piTmpSrcJudge, Int d, Int d0, Int  d3, Int dp, Int dq, Int beta, Int tc,  Int iThrCut, Int iSideThr, Int iBitdepthScale)
{   
  if (d>=beta)
  {
    return ;
  }

  Bool bFilterP = 0;
  Bool bFilterQ = 0;

  if ( dp <iSideThr)
  {
    bFilterP = 1;
  }
  if ( dq <iSideThr)
  {
    bFilterQ = 1;
  }

  if(  xUseStrongFiltering( iOffset, 2*d0, beta, tc , piTmpSrc+iSrcStep*(iShift+0))
    && xUseStrongFiltering( iOffset, 2*d3, beta, tc , piTmpSrc+iSrcStep*(iShift+3)))
  {
    return;
  }


  for ( UInt i = 0; i < DEBLOCK_SMALLEST_BLOCK/2; i++)
  {
    xPelFilterLumaSSDTc( piTmpSrc+iSrcStep*(iShift+i), iOffset, d, beta, tc , piTmpSrcJudge+iSrcStep*(iShift+i), iThrCut, bFilterP, bFilterQ, iBitdepthScale);
  }

}


__inline Void TEncLoopFilter::xPelFilterLumaSSDTc( Pel* piSrc, Int iOffset, Int d, Int beta, Int tc , Pel* piSrcJudge, Int iThrCut, Bool bFilterSecondP, Bool bFilterSecondQ, Int iBitdepthScale)
{
  Int delta,dif ;

  Pel m4  = piSrc[0];
  Pel m3  = piSrc[-iOffset];
  Pel m5  = piSrc[ iOffset];
  Pel m2  = piSrc[-iOffset*2];
  Pel m6  = piSrc[ iOffset*2];
  Pel m1  = piSrc[-iOffset*3];

  Pel m4j  = piSrcJudge[0];
  Pel m3j  = piSrcJudge[-iOffset];
  Pel m5j  = piSrcJudge[ iOffset];
  Pel m2j  = piSrcJudge[-iOffset*2];

  /* Weak filter */

  Int delta_full = (9*(m4-m3) -3*(m5-m2) + 8)>>4 ;

  if ( abs(delta_full) < iThrCut )
  {

    for(Int iCurTcOffset = -m_iTcRange;  iCurTcOffset< m_iTcRange +1 ; iCurTcOffset++)
    {
      Int iTc =  Clip3(0, 14*iBitdepthScale, (tc+iCurTcOffset * iBitdepthScale));

      Int SSD0 = 0;
      Int SSD1 = 0;

      delta = Clip3(-iTc, iTc, delta_full);        
      dif = m3j - m3 - delta;
      SSD0 += dif*dif; 
      dif = m4j - m4 + delta;
      SSD0 += dif*dif;

      m_iCounter[MAX_OFFSET+iCurTcOffset][0] += SSD0;       


      Int iTc2 = iTc>>1;

      if(bFilterSecondP)
      {
        Int delta1 = Clip3(-iTc2, iTc2, (( ((m1+m3+1)>>1)- m2+delta)>>1));
        dif = m2j - m2 - delta1;
        SSD1 += dif*dif; 
      }
      if(bFilterSecondQ)
      {
        Int delta2 = Clip3(-iTc2, iTc2, (( ((m6+m4+1)>>1)- m5-delta)>>1));
        dif = m5j - m5 - delta2;
        SSD1 += dif*dif; 

      }
      m_iCounter[MAX_OFFSET+iCurTcOffset][0] += SSD1;       
    }
  }
}


__inline Int TEncLoopFilter::xCalcThrCut( Pel* piSrc, Int iOffset)
{
  return abs ((9*(piSrc[0]-piSrc[-iOffset]) - 3*(piSrc[ iOffset]-piSrc[-iOffset*2]) + 8)>>4) ;
}

__inline Int TEncLoopFilter::xCalcDStrong( Pel* piSrc, Int iOffset)
{
  return abs( piSrc[-iOffset*4]-piSrc[-iOffset] ) + abs( piSrc[ iOffset*3] - piSrc[0] );

}
__inline Int TEncLoopFilter::xCalcStrongMiddle( Pel* piSrc, Int iOffset)
{
  return abs(piSrc[-iOffset]-piSrc[0]);
}

/**
- Deblocking for the luminance component with strong or weak filter
.
\param piSrcJudge    pointer to picture data for decision
\param bPartPNoFilter  indicator to disable filtering on partP
\param bPartQNoFilter  indicator to disable filtering on partQ
*/

__inline UInt TEncLoopFilter::xSSDPelOrig( Pel* piSrc, Pel* piOrig, Int iOffset )
{
  UInt uiSSDY = 0;
  for (Int i = -3; i< 3; i++)
  {
    Int iDiff = (Int)( piOrig[iOffset*i] - piSrc[iOffset*i]);
    uiSSDY   += iDiff * iDiff;    
  }
  return uiSSDY;
}

__inline UInt TEncLoopFilter::xSSDPelStrongFilt(  Pel* piSrc, Pel* piOrig, Int iOffset )
{
  Pel m4  = piSrc[0];
  Pel m3  = piSrc[-iOffset];
  Pel m5  = piSrc[ iOffset];
  Pel m2  = piSrc[-iOffset*2];
  Pel m6  = piSrc[ iOffset*2];
  Pel m1  = piSrc[-iOffset*3];
  Pel m7  = piSrc[ iOffset*3];
  Pel m0  = piSrc[-iOffset*4];

  UInt uiSSDY = 0;
  Int iDiff; 

  iDiff  = piOrig[-iOffset] - (( m1 + 2*m2 + 2*m3 + 2*m4 + m5 + 4) >> 3) ;
  uiSSDY   += iDiff * iDiff;    

  iDiff  = piOrig[0] - (( m2 + 2*m3 + 2*m4 + 2*m5 + m6 + 4) >> 3) ;
  uiSSDY   += iDiff * iDiff;    

  iDiff  = piOrig[-iOffset*2] - (( m1 + m2 + m3 + m4 + 2)>>2);
  uiSSDY   += iDiff * iDiff; 

  iDiff  = piOrig[ iOffset] - (( m3 + m4 + m5 + m6 + 2)>>2);
  uiSSDY   += iDiff * iDiff; 

  iDiff  = piOrig[-iOffset*3] - (( 2*m0 + 3*m1 + m2 + m3 + m4 + 4 )>>3);
  uiSSDY   += iDiff * iDiff; 

  iDiff  = piOrig[ iOffset*2] - (( m3 + m4 + m5 + 3*m6 + 2*m7 +4 )>>3);
  uiSSDY   += iDiff * iDiff; 

  return uiSSDY;

}

__inline UInt TEncLoopFilter::xSSDPelWeakFilt(  Pel* piSrc, Pel* piOrig, Int iOffset, UInt& ruiSSD2PelP, UInt& ruiSSD2PelQ, UInt& ruiSSDCenter, Int tc)
{
  Pel m4  = piSrc[0];
  Pel m3  = piSrc[-iOffset];
  Pel m5  = piSrc[ iOffset];
  Pel m2  = piSrc[-iOffset*2];
  Pel m6  = piSrc[ iOffset*2];
  Pel m1  = piSrc[-iOffset*3];



  UInt uiSSDY = 0;
  ruiSSD2PelP = ruiSSD2PelQ = ruiSSDCenter = 0;

  Int iDiff;
  UInt uiSqDiff;

  iDiff  = piOrig[-iOffset*3] - piSrc[-iOffset*3];
  uiSqDiff   = iDiff * iDiff;    
  uiSSDY += uiSqDiff;
  ruiSSD2PelP += uiSqDiff;
  ruiSSD2PelQ += uiSqDiff; 
  ruiSSDCenter += uiSqDiff;

  iDiff  = piOrig[iOffset*2] - piSrc[iOffset*2];
  uiSqDiff   = iDiff * iDiff;    
  uiSSDY += uiSqDiff;
  ruiSSD2PelP += uiSqDiff;
  ruiSSD2PelQ += uiSqDiff; 
  ruiSSDCenter += uiSqDiff;

  Int delta = Clip3(-tc, tc, (9*(m4-m3) -3*(m5-m2) + 8)>>4) ;

  iDiff = piOrig[-iOffset] - Clip((m3+delta));
  uiSqDiff   = iDiff * iDiff;    
  uiSSDY += uiSqDiff;
  ruiSSD2PelP += uiSqDiff;
  ruiSSD2PelQ += uiSqDiff; 
  ruiSSDCenter += uiSqDiff;

  iDiff = piOrig[0] - Clip((m4-delta));
  uiSqDiff  = iDiff * iDiff;    
  uiSSDY += uiSqDiff;
  ruiSSD2PelP += uiSqDiff;
  ruiSSD2PelQ += uiSqDiff; 
  ruiSSDCenter += uiSqDiff;

  Int tc2 = tc>>1;

  Int delta1 = Clip3(-tc2, tc2, (( ((m1+m3+1)>>1)- m2+delta)>>1));
  iDiff = piOrig[-iOffset*2] - Clip((m2+delta1));
  uiSqDiff  = iDiff * iDiff;    
  uiSSDY += uiSqDiff;
  ruiSSD2PelP += uiSqDiff;


  Int delta2 = Clip3(-tc2, tc2, (( ((m6+m4+1)>>1)- m5-delta)>>1));
  iDiff = piOrig[ iOffset] - Clip((m5+delta2));
  uiSqDiff  = iDiff * iDiff;    
  uiSSDY += uiSqDiff;
  ruiSSD2PelQ += uiSqDiff;

  iDiff = piOrig[ -iOffset*2] - piSrc[ -iOffset*2];
  uiSqDiff  = iDiff * iDiff;    
  ruiSSD2PelQ += uiSqDiff; 
  ruiSSDCenter += uiSqDiff;


  iDiff = piOrig[ iOffset] - piSrc[ iOffset];
  uiSqDiff  = iDiff * iDiff;    
  ruiSSD2PelP += uiSqDiff;
  ruiSSDCenter += uiSqDiff;

  return uiSSDY;
}


Int TEncLoopFilter::findBetaOffset( Int iOffset, Int iQP, Int iDiv)
{
  Int i;
  Int iTargetBeta = Clip3(0, 64, betatable_8x8[Clip3(0,51,iQP)]+iOffset);
  if (iOffset < 0)
  {
    for (i = 0; i>-13; i--)
    {
      if (betatable_8x8[Clip3(0,52,iQP+i*iDiv)]<=iTargetBeta)
      {
        return i;
      }
    }
  }
  else
  {
    for (i = 0; i<13; i++)
    {
      if (betatable_8x8[Clip3(0,52,iQP+i*iDiv)]>=iTargetBeta)
      {
        return i;
      }
    }
  }
  return i;
}


Int TEncLoopFilter::findSideBetaOffset( Int iOffset, Int iQP, Int iDiv)
{
  Int i;
  Int iCurBeta = betatable_8x8[Clip3(0,51,iQP)];
  Int iTargetBeta = Clip3(0, 13,((iCurBeta+(iCurBeta>>1))>>3)+iOffset);
  Int iNewSideBeta;

  if (iOffset < 0)
  {
    for (i = 0; i>-13; i--)
    {
      iNewSideBeta = betatable_8x8[Clip3(0,52,iQP+i*iDiv)];
      if (((iNewSideBeta+(iNewSideBeta>>1))>>3)<=iTargetBeta)
      {
        return i;
      }
    }
  }
  else
  {
    for (i = 0; i<13; i++)
    {
      iNewSideBeta = betatable_8x8[Clip3(0,52,iQP+i*iDiv)];
      if (((iNewSideBeta+(iNewSideBeta>>1))>>3)>=iTargetBeta)
      {
        return i;
      }
    }
  }
  return i;
}

Int TEncLoopFilter::findTcOffset( Int iOffset,  Int iQP, Int iDiv)
{
  Int i;
  Int iTargetTc = Clip3(0, 14, tctable_8x8[Clip3(0,53,iQP)]+iOffset);
  if (iOffset < 0)
  {
    for (i = 0; i>-13; i--)
    {
      if (tctable_8x8[Clip3(0,53,iQP+i*iDiv)]<=iTargetTc)
      {
        return i;
      }
    }
  }
  else
  {
    for (i = 0; i<13; i++)
    {
      if (tctable_8x8[Clip3(0,53,iQP+i*iDiv)]>=iTargetTc)
      {
        return i;
      }
    }
  }
  return i;
}

#endif




